import {ComponentType} from 'react';
import {useState, useEffect, useRef} from 'react';
import {Route, Link, generatePath, matchPath, useHistory, useLocation} from 'react-router-dom';

const routes = {

}

export const addRoute = ({name, path, props}) => routes[name || path] = {...props, path}

export const NamedRoute = ({name, routeName, query, pendingQueryComponent:P, errorQueryComponent:E, errorQueryComponentErrorProp="error", componentQueryResultProp='data', path, component, render, ...props}) => {
    name  ||= routeName
    routes[name || path] = {...props, path}

    const Render:ComponentType = component || render 

    const ComponentWrapper = (props) => {
        const [queryResult, setQueryResult] = useState()
        const [queryError, setQueryError]   = useState()
        const queryPubUnsubscribeRef        = useRef()
        const paramsRef                     = useRef()

        paramsRef.current = props.match.params

        useEffect(() => {
            if (query && typeof query === 'function'){
                if (typeof query.getQueryPubBranch === 'function'){ 
                    query(paramsRef.current)
                    const [queryPub, key] = query.getQueryPubBranch(paramsRef.current)
                    const pubToState = () => {
                        //console.log(queryPub, key, queryPub[key], queryPub[key].invalidated)
                        if (!queryPub[key]){ 
                            if (queryError || queryResult){
                                setQueryError()
                                setQueryResult()
                            }
                            return
                        }
                        const {status, payload, error, invalidated} = queryPub[key];
                        //console.log(invalidated);
                        if (invalidated){
                            queryPubUnsubscribeRef.current()
                            query(paramsRef.current)
                            const [queryPub] = query.getQueryPubBranch(paramsRef.current)
                            //console.log(queryPub, Object.keys(queryPub))
                            queryPubUnsubscribeRef.current = queryPub.subscribe(pubToState)
                            setQueryError()
                            setQueryResult()
                            return;
                        }

                        ;({
                            PENDING(){
                                setQueryError()
                                setQueryResult()
                            },
                            FULFILLED(){
                                setQueryResult(payload)
                            },
                            REJECTED(){
                                setQueryError(error)
                            }
                        })[status]();
                    }
                    pubToState()
                    queryPubUnsubscribeRef.current = queryPub.subscribe(pubToState)
                }
                else {
                    const querySyncResult = query(props.match.params, props, name)
                    if (querySyncResult && typeof querySyncResult.then === 'function'){
                        querySyncResult.then(setQueryResult, setQueryError)
                    }
                    else {
                        setQueryResult(querySyncResult)
                    }
                }
            }
            return () => {
                //console.log('UNMOUNT', props.match.params)
                setQueryError()
                setQueryResult()
                if (queryPubUnsubscribeRef.current)
                    queryPubUnsubscribeRef.current()
            }
        },Object.values(props.match.params))

        //console.log('NAMED ROUTE STATE', queryResult, queryError)

        if (!query || typeof query !== 'function')
            return <Render {...{[componentQueryResultProp]:query, ...props}} />

        if (!queryResult && !queryError)
            return P ? <P {...props} /> : null

        if (queryError)
            return E ? <E {...{[errorQueryComponentErrorProp]: queryError, ...props}} /> : null

        return <Render {...{[componentQueryResultProp]:queryResult, ...props}} />
    }

    return <Route {...props} path={path}  component={ComponentWrapper}/>
}

export const NamedLink = ({routeName, name, params, to,...props}) => {
    name  ||= routeName
    const route = routes[name]
    if (route && !to){
        const pattern = route.path
        to            = generatePath(pattern, params)
    }
    else {
        throw new Error(`Route "${name}"not registered`)
        //console.log('no route, hnyk', routes, to, props)
    }
    return <Link to={to} {...props} />
}

export const HashRoute = ({component, render, ...props}) => {
    const history = useHistory()
    const location = useLocation()
    const [match, setMatch] = useState<null|object>(null)
    const checkMatch = () => {
        setMatch(matchPath(window.location.hash.slice(1), props))
    }
    useEffect(checkMatch,[props.path, props.exact, props.strict, props.sensitive])
    useEffect(() => {
        window.addEventListener('hashchange', checkMatch)
        return () => {
            window.removeEventListener('hashchange', checkMatch)
        }
    },[])
    if (!match) return null

    const Render:ComponentType = component || render 

    return <Render match={match} history={history} location={location} />
}

export type GetUserRolesFunc = () => Array<string>

export const createPrivateRoute = (getUserRoles:GetUserRolesFunc, defaultFallback:string|Function, defaultRouteRoles:string[]):Function => 
    (RouteComponent:ComponentType=NamedRoute):ComponentType => 
        ({fallback, roles=[], render, component, ...props}) => {
            const Render:ComponentType = component || render 

            const intersectArrays = (first:Array<any>, second:Array<any>) => 
                first.find(firstItem => second.find(secondItem => firstItem === secondItem))

            const ComponentWrapper:ComponentType = (props) => {
                fallback               ||= defaultFallback
                roles:Array<string>    ||= defaultRouteRoles

                const userRoles:Array<string> = getUserRoles() || []

                if (!fallback){
                    throw new ReferenceError('fallback or defaultFallback are mandatory for Private Route')
                }

                //console.log('PRIVATE ROUTING', roles, userRoles)

                if (!intersectArrays(roles, userRoles)){
                    if (typeof fallback === 'string')
                        props.history.push(fallback)
                    if (typeof fallback === 'function')
                        fallback(props.history, props)
                    return null;
                }


                return <Render {...props} />
            }

            if (!getUserRoles || typeof getUserRoles !== 'function'){
                throw new ReferenceError('getUserRoles are mandatory for Private Routing and it should be a function')
            }


            return <RouteComponent {...props} component={ComponentWrapper}/>
        }

//console.log(createPrivateRoute)

    

//export const HashNamedRoute = ({name, routeName, query, pendingQueryComponent:P, errorQueryComponent:E, errorQueryComponentErrorProp="error", componentQueryResultProp='data', path, component, render, ...props}) => {


